import logging
import time
from io import StringIO

from nicegui import ui
from rich.console import Console
from rich.progress import track

file = StringIO()
console = Console(file=file, log_path=False)
console.print("Hello, world!")
logger = logging.getLogger()


class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            self.element.push(msg)
        except Exception:
            self.handleError(record)


log = ui.log(max_lines=10).classes("w-full")


async def update_log():
    print("Updating logs...")
    file.seek(0)
    content = file.read()
    log.push(content)


async def progress_test():
    total = 0
    for _ in track(range(100), description="Processing...", console=console, auto_refresh=True):
        # Fake processing time
        time.sleep(0.1)
        total += 1
    print(f"Processed {total} things.")


# ui.timer(0.5, update_log)


def update_label(x):
    print(x)
    file.seek(0)
    return file.read()


handler = LogElementHandler(log)
ui.label().bind_text_from(file, backward=update_label)
logger.addHandler(handler)
ui.context.client.on_disconnect(lambda: logger.removeHandler(handler))
ui.button("Log time", on_click=progress_test)

ui.run(port=9200)



import asyncio
import logging
import re
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable

import pandas as pd
from nicegui import run, ui
from nicegui.events import ValueChangeEventArguments

from wikibasemigrator import config
from wikibasemigrator.model.profile import WikibaseMigrationProfile
from wikibasemigrator.wikibase import Query

logger = logging.getLogger(__name__)


class WikibaseItemSelector(ABC):
    """
    abstract class the defining the wikibase entity selection interface
    """

    def __init__(self, profile: WikibaseMigrationProfile, selection_callback: Callable[[list[str]], Awaitable[None]]):
        self.profile = profile
        self.selection_callback = selection_callback

    @abstractmethod
    def setup_ui(self):
        """
        setup the ui of the element
        """
        raise NotImplementedError

    @abstractmethod
    def get_selected_entities(self) -> list[str]:
        """
        Get selected entities
        :return: list of the selected entities
        """
        raise NotImplementedError

    async def handle_selection_callback(self):
        """
        Handle the selection callback
        :param selected_ids: list of selected entities
        """
        await self.selection_callback(self.get_selected_entities())


class ItemSelectorElement(WikibaseItemSelector):
    """
    Select the
    """

    SEPERATOR = ","

    def __init__(self, profile: WikibaseMigrationProfile, selection_callback: Callable[[list[str]], Awaitable[None]]):
        super().__init__(profile, selection_callback)
        self.value: str = ""
        self.rows: list[dict] = []
        self.selection_display: ui.table | None = None
        self._worker: asyncio.Task | None = None

    def setup_ui(self):
        """
        setup the view
        """
        with ui.element("div") as container:
            container.classes("flex flex-col gap-2")
            ui.label("Provide a list of entities to translate. The list of entities must be comma seperated.")
            entity_input = ui.input(
                label="Item IDs",
                placeholder="e.g. Q1,Q123,Q423,...",
                on_change=self._handle_value_change,
                validation={"InvalidFormat": self._validate_value},
            )
            entity_input.classes("w-full rounded-md  hover:border-blue-500 border-blue-200")
            entity_input.props("clearable").props("outlined")
            entity_input.bind_value(self)
            entity_input.on("keydown.enter", self.handle_selection_callback)
            ui.button("Translate", on_click=self.handle_selection_callback)

            ui.label("Selected entities:")
            with ui.element(tag="div"):
                columns = [
                    {"name": "qid", "label": "Qid", "field": "qid", "required": True, "align": "left"},
                    {"name": "label", "label": "Label", "field": "label", "sortable": True, "align": "left"},
                ]

                self.selection_display = ui.table(columns=columns, rows=self.rows, row_key="qid")
                self.selection_display.add_slot(
                    "body-cell-qid",
                    f"""
                    <q-td :props="props">
                        <a :href="'{self.profile.source.item_prefix}' + props.value">{{{{ props.value }}}}</a>
                    </q-td>
                """,
                )

    def value_is_valid(self) -> bool:
        """
        Check if the value string is valid
        :return: True if the value string is valid, False otherwise
        """
        return self._validate_value(self.value)

    def get_selected_entities(self) -> list[str]:
        if self.value_is_valid():
            return self.value.strip().split(self.SEPERATOR)
        else:
            return []

    def _validate_value(self, value: str) -> bool:
        """
        Validate the input value
        :param value:
        :return:
        """
        if value is None:
            return True
        pattern = re.compile("^([PQ][0-9]+)(,([PQ][0-9]+))*$")
        return bool(pattern.match(value))

    async def _handle_value_change(self, event: ValueChangeEventArguments) -> None:
        if self.selection_display is None:
            logger.debug("Selection display is not setup → Abort lookup and display of labels")
            return
        if self.value and self._validate_value(self.value):
            if self._worker and not self._worker.done():
                self._worker.cancel()
            values = self.value.split(self.SEPERATOR)
            worker = asyncio.create_task(
                run.io_bound(
                    callback=Query.get_entity_label,
                    entity_ids=values,
                    endpoint_url=self.profile.source.sparql_url,
                    item_prefix=self.profile.source.item_prefix,
                )
            )
            self._worker = worker
            response = await worker
            if response:
                self.rows = response
            else:
                self.rows = []
            self.selection_display.update_rows(self.rows)


class QueryItemSelectorElement(WikibaseItemSelector):
    """
    Allows to select entities by providing a query
    """

    def __init__(self, profile: WikibaseMigrationProfile, selection_callback: Callable[[list[str]], Awaitable[None]]):
        """
        constructor
        :param profile: migration profile
        """
        super().__init__(profile, selection_callback)
        self.value = ""
        self.result_container: ui.element | None = None
        self.result: list[dict] | None = None

    def setup_ui(self):
        """
        setup the view
        """
        with ui.element("div") as container:
            container.classes("flex flex-col gap-2")
            ui.label(
                f"The SPARQL query must bind the ID of the entities to translate to the variable ?{config.ITEM_QUERY_VARIABLE}."  # noqa: E501
            )
            with ui.element("div") as div:
                div.classes("w-full h-min-1/2")
                editor = ui.codemirror(language="SPARQL", theme="githubLight")
                editor.classes("ring-2 ring-offset-2 border-blue-200 rounded-md")
                editor.bind_value(self)
            ui.button("Query", on_click=self._handle_query)

            self.result_container = ui.element(tag="div").classes("flex flex-col gap-2")
            self.display_result()

    async def _handle_query(self) -> None:
        """
        handle query b< executing the query and displaying the result
        :return:
        """
        if self.result_container is None:
            logger.debug("Result container is not setup → aborting handling of query")
            return
        self.result_container.clear()
        with self.result_container:
            ui.spinner(size="lg").classes("mx-auto")
        response = await run.io_bound(
            callback=Query.execute_query,
            query=self.value,
            endpoint_url=self.profile.source.sparql_url,
        )

        if response:
            self.result = response
            self.display_result()
        else:
            ui.notify("Query has no results", color="warning")

    def get_selected_entities(self) -> list[str]:
        """
        Get selected entities
        """
        if self.result is None:
            return []
        entity_ids_with_none: list[str | None] = [record.get(config.ITEM_QUERY_VARIABLE) for record in self.result]
        entity_ids_with_prefix: list[str] = [entity_id for entity_id in entity_ids_with_none if entity_id is not None]
        entity_ids = [
            entity_id.removeprefix(self.profile.source.item_prefix.unicode_string())
            for entity_id in entity_ids_with_prefix
        ]
        return entity_ids

    def display_result(self) -> None:
        """
        display the result of the query
        """
        if self.result and self.result_container is not None:
            self.result_container.clear()
            with self.result_container:
                df = pd.DataFrame.from_records(self.result)
                ui.aggrid.from_pandas(df).classes()
                valid_result = True
                with ui.element(tag="div").classes("flex flex-row gap-2"):
                    ui.label(f"{len(self.result)} results").classes("bg-green rounded-full px-2 py-1 flex-none")
                    if config.ITEM_QUERY_VARIABLE not in df.columns:
                        ui.label(f"?{config.ITEM_QUERY_VARIABLE} variable missing").classes(
                            "bg-red rounded-full px-2 py-1 flex-none"
                        )
                        valid_result = False
                translate_btn = ui.button(
                    "Translate",
                    on_click=self.handle_selection_callback,
                )
                translate_btn.set_enabled(valid_result)



import unittest

from wikibaseintegrator import WikibaseIntegrator

from wikibasemigrator.qsgenerator import QuickStatementsGenerator


class TestQuickStatementsGenerator(unittest.TestCase):
    def test_generation(self):
        """
        Q77966361
        :return:
        """
        wbi = WikibaseIntegrator()
        # Q77966361
        entity = wbi.item.get("Q80", mediawiki_api_url="https://www.wikidata.org/w/api.php")
        qs_generator = QuickStatementsGenerator()
        qs = qs_generator.generate_item(entity)
        print(qs)

    def test_generation_of_new_item(self):
        """
        test generation of new item
        """
        wbi = WikibaseIntegrator()
        new_item = wbi.item.new()
        new_item.labels.set("en", "New item")
        new_item.labels.set("fr", "Nouvel élément")
        qs_generator = QuickStatementsGenerator()
        qs = qs_generator.generate_item(new_item)
        expected_qs = """CREATE\nLAST|Len|"New item"\nLAST|Lfr|"Nouvel élément\""""
        self.assertEqual(qs, expected_qs)


if __name__ == "__main__":
    unittest.main()



import logging
from concurrent.futures import Future

from nicegui import run, ui

from wikibasemigrator.migrator import WikibaseMigrator
from wikibasemigrator.model.translations import EntitySetTranslationResult
from wikibasemigrator.web.components.progress import ProgressBar

logger = logging.getLogger(__name__)


class MigrationView:
    """
    Provides a progress overview
    """

    def __init__(self, migrator: WikibaseMigrator):
        """
        constructor
        :param profile: Migration profile
        """
        self.migrator = migrator
        self.profile = migrator.profile
        self.container: ui.element | None = None
        self.migration_container: ui.element | None = None
        self.progress_counter = 0
        self.progress_bar: ProgressBar | None = None
        self.migration_log: ui.log | None = None

    def setup_ui(self):
        with ui.element("div").classes("container mx-auto flex flex-col") as self.container:
            ui.label("Migrating Entities").classes("text-xl")
            with ui.element("div").classes("container flex flex-col gap-2") as self.migration_container:
                ui.spinner()

    async def start_migration(self, translations: EntitySetTranslationResult, summary: str):
        if self.migration_container is None:
            logger.debug("Migration container has not been setup")
            return
        ui.notify("Staring migration")
        self.migration_container.clear()
        with self.migration_container:
            self.progress_bar = ProgressBar(total=len(translations.get_target_entities()))
            self.migration_log = ui.log(max_lines=10).classes("w-full h-90")

        await run.io_bound(
            self.migrator.migrate_entities_to_target,
            translations,
            summary=summary,
            entity_done_callback=self._update_progress,
        )
        ui.notify("Migration completed", type="positive")
        self.migration_container.clear()
        with self.migration_container:
            with ui.link(target="/"):
                ui.button("Back to selection")
            self.display_entities(translations)

    def _update_progress(self, future: Future):
        """
        Update the progress bar
        :param future:
        :return:
        """
        result = future.result()
        if self.progress_bar:
            self.progress_bar.increment()
        if self.migration_log:
            if result.created_entity is None:
                self.migration_log.push(f"Migration of entity {result.original_entity.id} failed")
            else:
                self.migration_log.push(f"Migrated entity: {result.created_entity.id}")

    def display_entities(self, translations: EntitySetTranslationResult):
        """
        Display the given entities as table
        :param entities:
        :return:
        """
        rows = []
        rows_errors = []
        for translation in translations:
            if translation.created_entity is None:
                logger.debug("created_entity is not defined for translation → skipping")
                label = (
                    translation.original_entity.labels.get("en").value
                    if "en" in translation.original_entity.labels.values
                    else None
                )
                url = f"{self.profile.source.item_prefix}{translation.original_entity.id}"
                rows_errors.append(
                    {
                        "link": f"""<a href="{url}" target="_blank">{translation.original_entity.id} ({label})</a>""",
                        "errors": str(translation.errors),
                    }
                )
            else:
                label = (
                    translation.created_entity.labels.get("en").value
                    if "en" in translation.created_entity.labels.values
                    else None
                )
                url = f"{self.profile.target.item_prefix}{translation.created_entity.id}"
                rows.append(
                    {
                        "id": translation.created_entity.id,
                        "label": label,
                        "link": f"""<a href="{url}" target="_blank">{url}</a>""",
                        "errors": str(translation.errors),
                    }
                )
        ui.label("Migrated Entities").classes("mx-auto")
        ui.aggrid(
            {
                "columnDefs": [
                    {"headerName": "ID", "field": "id"},
                    {"headerName": "Label", "field": "label"},
                    {"headerName": "URL", "field": "link"},
                    {
                        "headerName": "Migration Details",
                        "field": "errors",
                        "wrapText": True,
                        "autoHeight": True,
                    },
                ],
                "rowData": rows,
            },
            html_columns=[2],
        )
        if rows_errors:
            ui.label("Failed Migrations").classes("mx-auto")
            ui.aggrid(
                {
                    "columnDefs": [
                        {"headerName": "Source", "field": "link"},
                        {
                            "headerName": "Error Messages",
                            "field": "errors",
                            "wrapText": True,
                            "autoHeight": True,
                        },
                    ],
                    "rowData": rows_errors,
                },
                html_columns=[0],
            )
        # ToDo: uncomment for darkmode support
        # grid.classes(add="ag-theme-alpine-auto-dark", remove="ag-theme-balham ag-theme-balham-dark")



import logging
import urllib.parse
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Union

from pydantic import HttpUrl
from wikibaseintegrator.entities import ItemEntity, PropertyEntity
from wikibaseintegrator.models import Qualifiers, References, Snak
from wikibaseintegrator.wbi_enums import WikibaseSnakType

from wikibasemigrator.model.quickstatements import (
    CreateLine,
    CreatePropertyLine,
    DateLine,
    DateQualifier,
    EntityLine,
    EntityQualifier,
    MonolingualTextLine,
    MonolingualTextQualifier,
    QuantityLine,
    QuantityQualifier,
    TextLine,
    TextQualifier,
    TimeQualifier,
    render_lines,
)
from wikibasemigrator.wikibase import WikibaseEntityTypes

logger = logging.getLogger(__name__)


class QuickStatementsGenerator:
    """
    Generate quick statements based on a given ItemEntity
    """

    def get_qs_url(self, item: ItemEntity, url: HttpUrl) -> HttpUrl:
        """
        The QuickStatements API does not work as expected.
        Not recommended to use at the moment
        :param item:
        :param url:
        :return:
        """

        qs_line = self.generate_item(item, newline="||")
        query = urllib.parse.urlencode({"v1": qs_line})
        qs_url = HttpUrl(f"{url}#/{query}")
        return qs_url

    def generate_items(self, items: list[ItemEntity], newline: str = "\n") -> str:
        """
        Generate QuickStatements for given list of items.
        :param items:
        :param newline:
        :return:
        """
        with ThreadPoolExecutor() as executor:
            func = partial(self.generate_item, newline=newline)
            result = executor.map(func, items)
        item_seperator = newline + newline
        quickstatements = item_seperator.join(result)
        return quickstatements

    def generate_item(self, item: ItemEntity, newline: str = "\n") -> str:
        """
        generate quick statements
        """
        lines: list[EntityLine | CreateLine | TextLine | CreatePropertyLine] = []
        if item.id is None:
            if isinstance(item, ItemEntity):
                lines.append(CreateLine())
            elif isinstance(item, PropertyEntity):
                lines.append(CreatePropertyLine(datatype=item.datatype))
            else:
                logger.debug(f"Quickstatements does not support creation of entities of type {item.ETYPE}")
                return ""
        lines.extend(self._get_label_lines(item))
        lines.extend(self._get_description_lines(item))
        lines.extend(self._get_aliases_lines(item))
        if item.ETYPE in WikibaseEntityTypes.support_sitelinks():
            lines.extend(self._get_sidelink_lines(item))
        lines.extend(self._get_statement_lines(item))
        # ToDo: ADD support for lexemes
        return render_lines(lines, newline=newline)

    def _get_item_subject(self, item: ItemEntity) -> str:
        qid = item.id
        if qid is None:
            qid = "LAST"
        return qid

    def _get_label_lines(self, item: ItemEntity) -> list[TextLine]:
        """
        Convert the labels of the given item into quick statement lines
        """
        subject = self._get_item_subject(item)
        lines: list[TextLine] = []
        for label in item.labels:
            try:
                line = TextLine(subject=subject, predicate=f"L{label.language}", target=label.value)
                lines.append(line)
            except Exception as ex:
                logger.debug(f"Exception while converting label {label}")
                logger.error(ex)
        return lines

    def _get_sidelink_lines(self, item: ItemEntity) -> list[TextLine]:
        """
        Convert the sidelinks of the given item into quick statement lines
        """
        subject = self._get_item_subject(item)
        lines: list[TextLine] = []
        for sidelink in item.sitelinks.sitelinks.values():
            try:
                line = TextLine(subject=subject, predicate=f"S{sidelink.site}", target=sidelink.title)
                lines.append(line)
            except Exception as ex:
                logger.debug("Exception while converting label lines: %s", ex)
        return lines

    def _get_description_lines(self, item: ItemEntity) -> list[TextLine]:
        """
        Convert the description of the given item into quick statement lines
        """
        subject = self._get_item_subject(item)
        lines: list[TextLine] = []
        for description in item.descriptions:
            try:
                line = TextLine(subject=subject, predicate=f"D{description.language}", target=description.value)
                lines.append(line)
            except Exception as ex:
                logger.debug("Exception while converting description lines: %s", ex)
        return lines

    def _get_aliases_lines(self, item: ItemEntity) -> list[TextLine]:
        """
        Convert the aliases of the given item into quick statement lines
        """
        subject = self._get_item_subject(item)
        lines: list[TextLine] = []
        for aliases in item.aliases.aliases.values():
            for alias in aliases:
                try:
                    line = TextLine(subject=subject, predicate=f"A{alias.language}", target=alias.value)
                    lines.append(line)
                except Exception as ex:
                    logger.debug("Exception while converting alias lines: %s", ex)
        return lines

    def _get_statement_lines(self, item: ItemEntity) -> list[TextLine | DateLine | EntityLine]:
        """
        Convert the statements of the given item into quick statement lines
        """
        subject = self._get_item_subject(item)
        lines: list[TextLine] = []
        for claim in item.claims:
            predicate = claim.mainsnak.property_number
            datatype = claim.mainsnak.datatype
            qualifiers = self._get_statement_qualifiers(claim.qualifiers)
            references = self._get_statement_references(claim.references)
            qualifiers.extend(references)
            if claim.mainsnak.snaktype is not WikibaseSnakType.KNOWN_VALUE:
                continue
            if datatype == "wikibase-item":
                line = EntityLine(
                    subject=subject,
                    predicate=predicate,
                    target=claim.mainsnak.datavalue["value"]["id"],
                    qualifiers=qualifiers,
                )
            elif datatype in ["external-id", "string", "url", "commonsMedia"]:
                line = TextLine(
                    subject=subject,
                    predicate=predicate,
                    target=claim.mainsnak.datavalue["value"],
                    qualifiers=qualifiers,
                )
            elif datatype == "time":
                date_str = claim.mainsnak.datavalue["value"]["time"]
                precision = claim.mainsnak.datavalue["value"]["precision"]
                calendar = claim.mainsnak.datavalue["value"]["calendarmodel"]
                line = DateLine(
                    subject=subject,
                    predicate=predicate,
                    target=date_str,
                    calendar=calendar,
                    qualifiers=qualifiers,
                    precision=precision,
                )
            elif datatype == "monolingualtext":
                text = claim.mainsnak.datavalue["value"]["text"]
                language = claim.mainsnak.datavalue["value"]["language"]
                line = MonolingualTextLine(
                    subject=subject, predicate=predicate, target=text, language=language, qualifiers=qualifiers
                )
            elif datatype == "quantity":
                snak = claim.mainsnak
                if not snak.datavalue:
                    logger.debug("Skipping qualifier with 'unknown value'")
                    continue
                amount = snak.datavalue["value"]["amount"]
                unit = snak.datavalue["value"]["unit"]
                if unit == "1":
                    unit = None
                else:
                    unit = unit.split("/")[-1]
                tolerance = snak.datavalue["value"].get("tolerance", None)
                line = QuantityLine(subject=subject, predicate=predicate, target=amount, tolerance=tolerance, unit=unit)
            else:
                print(datatype)
                # ToDo: Coordiantes
                continue
            lines.append(line)
        return lines

    def _get_statement_qualifiers(self, qualifiers: Qualifiers):
        line_qualifiers = []
        for qualifier in qualifiers:
            line = self._convert_snak(qualifier)
            if line:
                line_qualifiers.append(line)
        return line_qualifiers

    def _get_statement_references(self, references: References):
        line_reference = []
        for reference_group in references:
            for i, reference in enumerate(reference_group):
                new_reference_group = i == 0
                line = self._convert_snak(reference, as_reference=True, new_reference_group=new_reference_group)
                if line:
                    line_reference.append(line)
        return line_reference

    def _convert_snak(
        self, snak: Snak, as_reference: bool = False, new_reference_group: bool = False
    ) -> Union[TextQualifier, DateQualifier, "QuantityQualifier", EntityQualifier, None]:
        if snak.snaktype is not WikibaseSnakType.KNOWN_VALUE:
            # other snaktypes are not supported by quick statements
            return None
        property_id = snak.property_number
        if as_reference:
            property_id = property_id.replace("P", "S")
            if new_reference_group:
                property_id = f"!{property_id}"
        datatype = snak.datatype
        if datatype in ["string", "url", "external-id"]:
            value = snak.datavalue["value"]
            line = TextQualifier(predicate=property_id, target=value)
        elif datatype == "time":
            date_str = snak.datavalue["value"]["time"]
            precision = snak.datavalue["value"]["precision"]
            calendar = snak.datavalue["value"]["calendarmodel"]
            line = TimeQualifier(predicate=property_id, target=date_str, precision=precision, calendar=calendar)
        elif datatype == "wikibase-item":
            line = EntityQualifier(predicate=property_id, target=snak.datavalue["value"]["id"])
        elif datatype == "monolingualtext":
            text = snak.datavalue["value"]["text"]
            language = snak.datavalue["value"]["language"]
            line = MonolingualTextQualifier(predicate=property_id, target=text, language=language)
        elif datatype == "quantity":
            if not snak.datavalue:
                logger.debug("Skipping qualifier with 'unknown value'")
                return None
            amount = snak.datavalue["value"]["amount"]
            unit = snak.datavalue["value"]["unit"]
            if unit == "1":
                unit = None
            else:
                unit = unit.split("/")[-1]
            tolerance = snak.datavalue["value"].get("tolerance", None)
            line = QuantityQualifier(predicate=property_id, target=amount, unit=unit, tolerance=tolerance)
        else:
            logger.debug(f"Skipping qualifier with datatype {datatype}")
            return None
        return line



import importlib.metadata

from wikibaseintegrator.entities import ItemEntity, LexemeEntity, MediaInfoEntity, PropertyEntity

__version__ = importlib.metadata.version("wikibasemigrator")

# Wikibaseintegrator BaseEntity does not support all the methods that the subclasses all have
WbEntity = ItemEntity | PropertyEntity | LexemeEntity | MediaInfoEntity



import unittest
from pathlib import Path

from wikibasemigrator.mapper import WikibaseItemMapper
from wikibasemigrator.model.profile import load_profile
from wikibasemigrator.wikibase import WbiDataTypes


class TestWikibaseItemMapper(unittest.TestCase):
    """
    Test WikibaseItemMapper
    """

    def setUp(self):
        self.profile_path = Path(__file__).parent.joinpath("../src/wikibasemigrator/profiles/FactGrid.yaml")
        self.profile = load_profile(self.profile_path)

    def test_querying_mapping(self):
        mapper = WikibaseItemMapper(self.profile)
        mapping = mapper.get_mapping_for("Q183")
        self.assertEqual("Q140530", mapping)
        self.assertEqual("P2", mapper.get_mapping_for("P31"))

    def test_caching_of_property_types(self):
        """ "
        test caching of property types
        """
        mapper = WikibaseItemMapper(self.profile)
        mapper.query_mappings_for(["P31", "P580", "Q5", "Q80"])
        self.assertIn("P31", mapper.source_property_types)
        self.assertIn("P580", mapper.source_property_types)
        self.assertIn("P2", mapper.target_property_types)
        self.assertEqual(WbiDataTypes.WIKIBASE_ITEM, mapper.source_property_types.get("P31"))
        self.assertEqual(WbiDataTypes.WIKIBASE_ITEM, mapper.target_property_types.get("P2"))



import json
from pathlib import Path
from unittest import TestCase

from deepdiff import DeepDiff
from wikibaseintegrator import WikibaseIntegrator
from wikibaseintegrator.entities import ItemEntity

from wikibasemigrator.merger import EntityMerger


class TestEntityMerger(TestCase):
    """
    Test EntityMerger
    """

    def test_merge(self):
        """
        test merging entities
        """
        self.maxDiff = None
        resource_dir = Path(__file__).parent / "resources/merger"
        source_json = json.loads(resource_dir.joinpath("Q80_source.json").read_text())["entities"]["Q80"]
        target_json = json.loads(resource_dir.joinpath("Q80_target.json").read_text())["entities"]["Q80"]
        expected_full_json = json.loads(resource_dir.joinpath("Q80_expected.json").read_text())["entities"]["Q80"]
        wbi = WikibaseIntegrator()
        source_entity = ItemEntity(api=wbi).from_json(source_json)
        target_entity = ItemEntity(api=wbi).from_json(target_json)
        expected_entity = ItemEntity(api=wbi).from_json(expected_full_json)
        merged_entity = EntityMerger().merge(source_entity, target_entity)
        actual_json = merged_entity.get_json()
        expected_json = expected_entity.get_json()
        result = DeepDiff(actual_json, expected_json, ignore_order=True)
        self.assertEqual(result, {})



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from string import Template

from wikibasemigrator import config, wikibase
from wikibasemigrator.model.profile import WikibaseConfig, WikibaseMigrationProfile
from wikibasemigrator.wikibase import WbiDataTypes

logger = logging.getLogger(__name__)


class WikibaseItemMapper:
    """
    Maps items from one wikibase instance to another
    """

    def __init__(self, profile: WikibaseMigrationProfile):
        self.migration_profile = profile
        self._raw_mappings: set[tuple[str, str]] = set()
        self.mappings: dict[str, str | None] = dict()
        self.source_property_types: dict[str, WbiDataTypes] = dict()
        self.target_property_types: dict[str, WbiDataTypes] = dict()

    @property
    def wikibase_config(self) -> WikibaseConfig:
        """
        Get wikibase config of the wikibase instance that holds the mapping information
        """
        return self.migration_profile.get_wikibase_config_of_mapping_location()

    def query_mappings_for(self, ids: list[str]):
        """
        query for the mappings of the given ids.
        The list of ids can be a mixture of Qids and Pids as they will be seperated and queried separately.
        :param ids: list of ids
        :return:
        """
        property_ids = [item for item in ids if item.startswith("P")]
        item_ids = [item for item in ids if item.startswith("Q")]
        self._init_cache_for(ids)
        ids_queries = [
            ("property", property_ids, self.migration_profile.mapping.property_mapping_query),
            ("item", item_ids, self.migration_profile.mapping.item_mapping_query),
        ]
        for id_type, id_values, query_raw in ids_queries:
            if id_values:
                query_template = Template(query_raw)
                source_items = [f'"{item}"' for item in id_values]
                logger.debug(f"Querying {id_type} mappings for {len(id_values)} IDs")
                lod = wikibase.Query.execute_values_query_in_chunks(
                    query_template=query_template,
                    param_name=config.MAPPING_QUERY_VALUES_INPUT_VARIABLE,
                    values=source_items,
                    endpoint_url=self.wikibase_config.sparql_url,
                )
                self._update_raw_cache(lod)
        self.update_property_type_map()
        self.update_cache()

    def update_property_type_map(self):
        """
        update the property type mappings for source and target properties
        """
        logger.debug("Updating property type mappings")
        with ThreadPoolExecutor(max_workers=2) as executor:
            executor.submit(self._update_source_property_types)
            executor.submit(self._update_target_property_types)

    def _update_source_property_types(self):
        """
        update the property type mappings for source properties
        :return:
        """
        logger.debug("Updating source property type mappings")
        used_props = {pid for pid, _ in self._raw_mappings if pid.startswith("P")}
        cached_types = set(self.source_property_types.keys())
        needed_props = used_props - cached_types
        prop_type_mappings = wikibase.Query.get_property_datatype(
            endpoint_url=self.migration_profile.source.sparql_url,
            property_ids=list(needed_props),
            item_prefix=self.migration_profile.source.item_prefix,
        )
        wbi_prop_type_mappings = {key: value.get_wbi_type() for key, value in prop_type_mappings.items()}
        self.source_property_types.update(wbi_prop_type_mappings)

    def _update_target_property_types(self):
        """
        update the property type mappings for target properties
        :return:
        """
        logger.debug("Updating target property type mappings")
        used_props = {pid for _, pid in self._raw_mappings if pid is not None and pid.startswith("P")}
        cached_types = set(self.target_property_types.keys())
        needed_props = used_props - cached_types
        prop_type_mappings = wikibase.Query.get_property_datatype(
            endpoint_url=self.migration_profile.target.sparql_url,
            property_ids=list(needed_props),
            item_prefix=self.migration_profile.target.item_prefix,
        )
        wbi_prop_type_mappings = {key: value.get_wbi_type() for key, value in prop_type_mappings.items()}
        self.target_property_types.update(wbi_prop_type_mappings)

    def query_mapping_for(self, item: str):
        """
        Query the mapping of the given item and update the cache
        """
        self.query_mappings_for(ids=[item])

    def _update_raw_cache(self, lod: list[dict]) -> None:
        """
        Update the cache
        :param lod: list of mappings to cache
        :return:
        """
        for record in lod:
            source = record.get(config.MAPPING_QUERY_SOURCE_VARIABLE, None)
            target = record.get(config.MAPPING_QUERY_TARGET_VARIABLE, None)
            if isinstance(source, str) and isinstance(target, str):
                self._raw_mappings.add((source, target))

    def update_cache(self):
        """
        update the mapping cache and ensure that
        if multiple mappings for the same property exist that the one with the correct type is chosen
        """
        d = {}
        for source, target in self._raw_mappings:
            d.setdefault(source, []).append(target)
        for source, target_ids in d.items():
            if len(target_ids) == 1:
                self.mappings[source] = target_ids[0]
            elif source.startswith("P"):
                source_type = self.source_property_types.get(source, None)
                matching_type = False
                for target_id in target_ids:
                    target_type = self.target_property_types.get(target_id, None)
                    if source_type == target_type:
                        logger.debug(
                            f"Source property {source} has multiple mappings → Choosing {target_id} from target as it has the same property datatype"  # noqa: E501
                        )
                        self.mappings[source] = target_id
                        matching_type = True
                        break
                if not matching_type:
                    logger.debug(
                        f"Property with multiple mappings for source {source} and targets {target_ids} → To Resolve this the first property of the list is chosen"  # noqa: E501
                    )
                    self.mappings[source] = target_ids[0]
            else:
                logger.debug(
                    f"Entity {source} has multiple mappings in target {target_ids} → To Resolve this the first property of the list is chosen"  # noqa: E501
                )
                self.mappings[source] = target_ids[0]

    def _init_cache_for(self, ids: list[str]) -> None:
        """
        Set cache to None for the given ids if not already cached
        :param ids:
        :return:
        """
        for id_value in ids:
            if id_value not in self.mappings:
                self.mappings[id_value] = None

    def prepare_cache_for(self, items: list[str]):
        """
        prepare the cache for a given list of items
        :param items: items  to cache the mapping for
        :return:
        """
        items_to_query = [item for item in items if item not in self.mappings]
        if items_to_query:
            logger.debug(
                f"Prepare cache for {len(items_to_query)} items. ({len(items) - len(items_to_query)} were already cached)"  # noqa: E501
            )
            start_time = datetime.now()
            self.query_mappings_for(items_to_query)
            logger.debug(f"Cache preparation took {datetime.now() - start_time}")

    def get_mapping_for(self, item: str) -> str | None:
        """
        Get the mapping for the given item in the target wikibase instance
        :param item: item id either a Qid or Pid
        :return: id of the corresponding item in the target wikibase instance
        """
        if not self.is_cached(item):
            self.query_mapping_for(item)
        return self.mappings.get(item, None)

    def is_cached(self, item: str) -> bool:
        """
        Check if the given entity id is cached
        :param item: entity id to check
        :return: True if it is cached, False otherwise
        """
        return item in self.mappings

    def get_existing_mappings(self) -> dict[str, str]:
        return {key: value for key, value in self.mappings.items() if value is not None}

    def get_missing_mappings(self) -> list[str]:
        return [key for key, value in self.mappings.items() if value is None]

    def get_missing_item_mapings(self) -> list[str]:
        return [value for value in self.get_missing_mappings() if value.startswith("Q")]

    def get_missing_property_mapings(self) -> list[str]:
        return [value for value in self.get_missing_mappings() if value.startswith("P")]



class UnknownEntityTypeException(Exception):
    """
    Type of the entity is not known.
    """

    def __init__(self, entity_id: str):
        self.entity_id = entity_id
        super().__init__(f'Unknown entity type "{entity_id}"')


class UserLoginRequiredException(Exception):
    """
    User login required.
    """



import logging
from collections.abc import Generator

from pydantic import BaseModel, ConfigDict, Field

from wikibasemigrator import WbEntity

logger = logging.getLogger(__name__)


class EntityTranslationResult(BaseModel):
    """
    wikibase migration translation result of an entity
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    entity: WbEntity
    original_entity: WbEntity
    missing_properties: list[str] = Field(default_factory=list)
    missing_items: list[str] = Field(default_factory=list)
    entity_mapping: dict[str, str | None] = Field(default_factory=dict)
    created_entity: WbEntity | None = None
    errors: list[str] = Field(default_factory=list)

    def add_missing_property(self, property_id: str):
        """
        add missing property
        :param property_id:
        :return:
        """

        self.missing_properties.append(property_id)

    def add_missing_item(self, entity_id: str):
        """
        add missing item
        :param entity_id:
        :return:
        """
        self.missing_items.append(entity_id)

    def add_entity_mapping(self, source_id: str, target_id: str | None):
        """
        add given mapping to mappings
        :param source_id:
        :param target_id:
        :return:
        """
        self.entity_mapping[source_id] = target_id

    def add_entity_mappings(self, mappings: dict[str, str | None]):
        """
        Update mappings with given mappings
        :param mappings:
        :return:
        """
        self.entity_mapping.update(mappings)


class EntitySetTranslationResult(BaseModel):
    entities: dict[str, EntityTranslationResult] = Field(default_factory=dict)

    @classmethod
    def from_list(cls, entities: list[EntityTranslationResult]) -> "EntitySetTranslationResult":
        res = EntitySetTranslationResult()
        for entity in entities:
            res.entities[entity.original_entity.id] = entity
        return res

    def get_created_entities(self) -> list[WbEntity]:
        return [entity.created_entity for entity in self.entities.values()]

    def get_missing_properties(self) -> list[str]:
        """
        Get IDs of missing properties
        :return: list of missing properties
        """
        missing = set()
        for entity in self.entities.values():
            missing.update(entity.missing_properties)
        return list(missing)

    def get_missing_items(self) -> list[str]:
        """
        Get IDs of missing item
        """
        missing = set()
        for entity in self.entities.values():
            missing.update(entity.missing_items)
        return list(missing)

    def get_mapping(self) -> dict[str, str | None]:
        """
        Get all mappings that are used
        :return:
        """
        mapping = dict()
        for entity in self.entities.values():
            mapping.update(entity.entity_mapping)
        return mapping

    def get_existing_mappings(self) -> dict[str, str]:
        """
        Get all existing mappings
        :return:
        """
        return {key: value for key, value in self.get_mapping().items() if value is not None}

    def get_missing_mappings(self) -> list[str]:
        """
        Get all missing mappings
        :return:
        """
        return [key for key, value in self.get_mapping().items() if value is None]

    def get_missing_item_mapings(self) -> list[str]:
        """
        Get all missing item mappings
        :return:
        """
        return [value for value in self.get_missing_mappings() if value.startswith("Q")]

    def get_missing_property_mapings(self) -> list[str]:
        """
        Get all missing property mappings
        :return:
        """
        return [value for value in self.get_missing_mappings() if value.startswith("P")]

    def get_translation_source_item_ids(self):
        """
        Get IDs of all main source items
        :return:
        """
        return [item.original_entity.id for item in self.entities.values()]

    def get_target_entities(self) -> list[WbEntity]:
        """
        Get all target items
        """
        return [translation_result.entity for translation_result in self.entities.values()]

    def get_source_entities(self) -> list[WbEntity]:
        """
        Get all source items
        """
        return [translation_result.original_entity for translation_result in self.entities.values()]

    def get_source_entity_ids(self) -> list[str]:
        """
        Get IDs of all source entities that are used
        """
        return [entity_id for entity_id in self.get_mapping()]

    def get_target_entity_ids(self) -> list[str]:
        """
        Get IDs of all target entities that are used
        """
        return [entity_id for entity_id in self.get_mapping().values() if entity_id is not None]

    def get_source_root_entity_ids(self):
        """
        Get IDs of all source items to migrate.
        Only the "subject" ids no ids used in any statement, qualifier or reference
        """
        return [item.id for item in self.get_source_entities() if item.id]

    def get_translation_result_by_source_id(self, source_id: str) -> EntityTranslationResult | None:
        """
        Get TranslationResult by source id
        :param source_id:
        :return:
        """
        for item in self.entities.values():
            if item.original_entity.id == source_id:
                return item
        return None

    def __iter__(self) -> Generator[EntityTranslationResult, None, None]:
        yield from self.entities.values()



import concurrent.futures
import logging
from pathlib import Path

from nicegui import app, ui

from wikibasemigrator import __version__
from wikibasemigrator.model.profile import WikibaseMigrationProfile
from wikibasemigrator.web.oauth import MediaWikiUserIdentity
from wikibasemigrator.wikibase import MediaWikiEndpoint, Query

logger = logging.getLogger(__name__)


class EndpointsAvailability:
    """
    Availability status of endpoints
    """

    source_sparql_endpoint: bool = False
    source_mediawiki_api: bool = False
    target_sparql_endpoint: bool = False
    target_mediawiki_api: bool = False

    def all_available(self):
        """
        Returns True if all endpoints are available
        :return:
        """
        return (
            self.source_sparql_endpoint
            and self.target_sparql_endpoint
            and self.target_mediawiki_api
            and self.source_mediawiki_api
        )


class Webpage:
    """
    Webpage template
    """

    def __init__(
        self,
        profile: WikibaseMigrationProfile,
        icon_path: Path | None = None,
        user: MediaWikiUserIdentity | None = None,
    ):
        self.profile = profile
        self.icon_path = icon_path
        self.user = user
        self.container: ui.element | None = None
        self.endpoints_availability = EndpointsAvailability()
        self.status_check: ui.timer | None = None

    def setup_ui(self):
        ui.colors(primary="#2c4e80ff")
        self.setup_header()
        self.setup_footer()
        self.container = ui.element(tag="div").classes("container flex flex-col mx-auto w-full h-full flex p-2")

    def setup_footer(self):
        """
        setup footer
        :return:
        """
        with ui.footer():
            with ui.element("div").classes("flex justify-between w-full"):
                with ui.element("div") as left:
                    left.classes("flex justify-start")
                    pass
                with ui.element("div") as middle:
                    middle.classes("flex justify-center")
                    ui.label(f"WikibaseMigrator {__version__}")
                with ui.element("div") as right:
                    right.classes("flex justify-end ")
                    self.status_container = ui.element("div").classes("flex flex-row")
                    self.status_check = ui.timer(0.1, self.check_service_availabilities)

    def setup_header(self):
        """
        setup page header
        """
        with ui.header().classes(replace="row items-center") as header:
            header.classes("bg-white dark:bg-slate-800 border-2 gap-2 flex flex-row p-2 gap-2")
            self.display_page_icon()
            with ui.link(target="/"):
                ui.button(icon="home", text="Home").props("flat")
            with ui.link(target="/config"):
                ui.button(icon="settings", text="Config").props("flat")
            with ui.element("a").classes(
                "flex flex-row text-lg text-black gap-1 hover:bg-slate-100 p-2 px-4 rounded"
            ) as link:
                link._props["href"] = "https://github.com/tholzheim/WikibaseMigrator"
                link._props["target"] = "_blank"
                ui.html(
                    content="""
                <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24">
                    <path d="M12 0c-6.626 0-12 5.373-12 12 0 5.302 3.438 9.8 8.207 11.387.599.111.793-.261.793-.577v-2.234c-3.338.726-4.033-1.416-4.033-1.416-.546-1.387-1.333-1.756-1.333-1.756-1.089-.745.083-.729.083-.729 1.205.084 1.839 1.237 1.839 1.237 1.07 1.834 2.807 1.304 3.492.997.107-.775.418-1.305.762-1.604-2.665-.305-5.467-1.334-5.467-5.931 0-1.311.469-2.381 1.236-3.221-.124-.303-.535-1.524.117-3.176 0 0 1.008-.322 3.301 1.23.957-.266 1.983-.399 3.003-.404 1.02.005 2.047.138 3.006.404 2.291-1.552 3.297-1.23 3.297-1.23.653 1.653.242 2.874.118 3.176.77.84 1.235 1.911 1.235 3.221 0 4.609-2.807 5.624-5.479 5.921.43.372.823 1.102.823 2.222v3.293c0 .319.192.694.801.576 4.765-1.589 8.199-6.086 8.199-11.386 0-6.627-5.373-12-12-12z">
                    </path>
                </svg>
                """  # noqa: E501
                )
                ui.label("GitHub")
            ui.element("div").classes("grow")
            if self.user:
                ui.button(self.user.username, on_click=self.logout).tooltip("Click to logout")
            else:
                oauth_login_url = "/login/wiki"
                with ui.link(target=oauth_login_url):
                    ui.button(icon="login", text="Login").props("flat")

    def display_page_icon(self):
        """
        Display page icon
        :return:
        """
        if self.icon_path.exists() and self.icon_path.is_file() and self.icon_path.suffix == ".svg":
            with ui.link(target="/"):
                ui.html(content=self.icon_path.read_text()).classes("w-32")

    def logout(self):
        app.storage.user["token"] = None
        app.storage.user["user"] = None
        ui.navigate.to("/", new_tab=False)

    def check_service_availabilities(self):
        """
        Check the availability of the used services
        """
        logger.info("Checking availability of used Services")
        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            future1 = executor.submit(Query.check_availability_of_sparql_endpoint, self.profile.source.sparql_url)
            future4 = executor.submit(MediaWikiEndpoint.check_availability, self.profile.source.mediawiki_api_url)
            future2 = executor.submit(Query.check_availability_of_sparql_endpoint, self.profile.target.sparql_url)
            future3 = executor.submit(MediaWikiEndpoint.check_availability, self.profile.target.mediawiki_api_url)

            self.endpoints_availability.source_sparql_endpoint = future1.result()
            self.endpoints_availability.target_sparql_endpoint = future2.result()
            self.endpoints_availability.target_mediawiki_api = future3.result()
            self.endpoints_availability.source_mediawiki_api = future4.result()

        self.setup_status_bar()

    def setup_status_bar(self) -> None:
        """
        Setup status bar showing the availability of the source and target services
        """
        icon_available = "wifi"
        icon_unavailable = "wifi_off"

        def show_endpoint_status(name: str, status: bool) -> None:
            with ui.element("div").classes("flex flex-row w-full"):
                ui.label(f"{name}:").classes("grow")
                ui.icon(icon_available if status else icon_unavailable, color="green" if status else "red").classes(
                    "justify-self-end"
                )

        if self.status_container is None:
            logger.error("Abort setup container not yet setup")
            return
        self.status_container.clear()

        with self.status_container:
            icon = icon_available if self.endpoints_availability.all_available() else icon_unavailable
            with ui.dialog() as dialog, ui.card().classes("flex flex-col"):
                ui.label("Endpoint availabilities").classes("text-xl")
                show_endpoint_status(
                    f"{self.profile.source.name} SPARQL Endpoint availability",
                    self.endpoints_availability.source_sparql_endpoint,
                )
                show_endpoint_status(
                    f"{self.profile.source.name} Mediawiki API availability",
                    self.endpoints_availability.source_mediawiki_api,
                )
                show_endpoint_status(
                    f"{self.profile.target.name} SPARQL Endpoint availability",
                    self.endpoints_availability.target_sparql_endpoint,
                )
                show_endpoint_status(
                    f"{self.profile.target.name} Mediawiki API availability",
                    self.endpoints_availability.target_mediawiki_api,
                )
                ui.button("Close", on_click=dialog.close)
            ui.icon(icon, color="green" if self.endpoints_availability.all_available() else "red").classes(
                "hover:cursor-pointer"
            ).on("click", dialog.open)
            if self.endpoints_availability.all_available() and self.status_check:
                self.status_check.deactivate()



import hashlib
import json
import logging
import tempfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from enum import Enum
from functools import partial
from pathlib import Path
from string import Template

import requests
from pydantic import HttpUrl
from SPARQLWrapper import JSON, POST, SPARQLWrapper
from wikibaseintegrator import __version__

logger = logging.getLogger(__name__)


WIKIBASE_PREFIX = "http://wikiba.se/ontology#"


def get_default_user_agent() -> str:
    """
    Get default user agent
    """
    return f"WikibaseMigrator/{__version__}"


class WbiDataTypes(str, Enum):
    """
    WikibaseIntegrator data types
    """

    STRING = "string"
    EXTERNAL_ID = "external-id"
    WIKIBASE_ITEM = "wikibase-item"
    TIME = "time"
    COMMONS_MEDIA = "commonsMedia"
    QUANTITY = "quantity"
    MONOLINGUALTEXT = "monolingualtext"
    GLOBE_COORDINATE = "globecoordinate"
    ENTITY_SCHEMA = "entity-schema"
    URL = "url"
    PROPERTY = "property"
    GEO_SHAPE = "geo-shape"
    TABUlAR_DATA = "tabulated-data"
    MATH = "math"
    SENSE = "wikibase-sense"
    MUSICAL_NOTATION = "musical-notation"
    LEXEME = "wikibase-lexeme"
    FORM = "wikibase-form"
    BASE_DATATYPE = "base-data-type"
    LOCAL_MEDIA = "localMedia"
    EDTF = "edtf"


class WikidataDataTypes(str, Enum):
    """
    wikidata data types
    """

    STRING = "String"
    EXTERNAL_ID = "ExternalId"
    WIKIBASE_ITEM = "WikibaseItem"
    TIME = "Time"
    COMMONS_MEDIA = "CommonsMedia"
    QUANTITY = "Quantity"
    MONOLINGUALTEXT = "Monolingualtext"
    GLOBE_COORDINATE = "GlobeCoordinate"
    ENTITY_SCHEMA = "EntitySchema"
    URL = "Url"
    PROPERTY = "WikibaseProperty"
    GEO_SHAPE = "GeoShape"
    TABUlAR_DATA = "TabularData"
    MATH = "Math"
    SENSE = "WikibaseSense"
    MUSICAL_NOTATION = "MusicalNotation"
    LEXEME = "WikibaseLexeme"
    FORM = "WikibaseForm"

    def get_wbi_type(self) -> WbiDataTypes:
        return WbiDataTypes[self.name]


class Query:
    """
    Holds basic functions to query a wikibase
    """

    @classmethod
    def get_entity_label(
        cls,
        endpoint_url: HttpUrl,
        entity_ids: list[str],
        item_prefix: HttpUrl,
        language: str | None = None,
    ) -> list[dict]:
        """
        Get the labels for the given entities
        :param endpoint_url:
        :param entity_ids:
        :param language: if None english will be used
        :param item_prefix:
        :return:
        """
        query_raw = Template("""
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        SELECT ?qid ?label
        WHERE{
          VALUES ?qid { 
            $entity_ids
          }
          BIND(IRI(CONCAT("$item_prefix", STR( ?qid))) as ?wd_qid)
          ?wd_qid rdfs:label ?label. FILTER(lang(?label)="$language")
        }
        """)
        if language is None:
            language = "en"
        query_template = Template(query_raw.safe_substitute(language=language, item_prefix=item_prefix))
        values = [f'"{entity_id}"' for entity_id in entity_ids]
        lod = cls.execute_values_query_in_chunks(
            query_template=query_template,
            param_name="entity_ids",
            values=values,
            endpoint_url=endpoint_url,
        )
        return lod

    @classmethod
    def get_property_datatype(
        cls, endpoint_url: HttpUrl, property_ids: list[str], item_prefix: HttpUrl
    ) -> dict[str, WikidataDataTypes]:
        """
        Get the datatype for the given list of property ids
        :param endpoint_url: endpoint to query
        :param property_ids: property ids for which the datatype should be returned
        :param item_prefix: namespace of the wikibase entities
        :return: mapping from the property ids to the datatype
        """
        query = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX wikibase: <http://wikiba.se/ontology#>
        SELECT * WHERE {
          VALUES ?p {
            $property_ids
          }
          ?p rdf:type wikibase:Property;
            wikibase:propertyType ?type.
        }
        """
        if not property_ids:
            return {}
        property_uris = map(partial(cls.add_namespace, namespace=item_prefix.unicode_string()), property_ids)
        values = list(map(cls.get_sparql_uri, property_uris))
        lod = cls.execute_values_query_in_chunks(
            query_template=Template(query),
            param_name="property_ids",
            values=values,
            endpoint_url=endpoint_url,
        )
        prop_type_map = dict()
        for d in lod:
            pid = cls.remove_namespace(d.get("p"), item_prefix.unicode_string())
            datatype = cls.remove_namespace(d.get("type"), WIKIBASE_PREFIX)
            prop_type_map[pid] = WikidataDataTypes(datatype)
        return prop_type_map

    @classmethod
    def chunks(cls, lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    @classmethod
    def execute_values_query_in_chunks(
        cls, query_template: Template, param_name: str, values: list[str], endpoint_url: HttpUrl, chunk_size: int = 1000
    ):
        """
        Execute given query in chunks to speedup execution
        :param chunk_size:
        :param endpoint_url:
        :param query_template:
        :param param_name:
        :param values:
        :return:
        """
        lod = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for item_id_chunk in cls.chunks(values, chunk_size):
                source_items = "\n".join(item_id_chunk)
                query = query_template.substitute(**{param_name: source_items})
                logger.debug(f"Querying chunk of size {len(item_id_chunk)} labels from {endpoint_url}")
                future = executor.submit(
                    cls.execute_query,
                    query=query,
                    endpoint_url=endpoint_url,
                )
                futures.append(future)
            for future in as_completed(futures):
                lod_chunk = future.result()
                lod.extend(lod_chunk)
        return lod

    @classmethod
    def execute_query(cls, query: str, endpoint_url: HttpUrl) -> list[dict]:
        """
        Execute given query against given endpoint
        :param query:
        :param endpoint_url:
        :return:
        """
        query_first_line = query.split("\n")[0][:30] if query.strip().startswith("#") else ""
        query_hash = hashlib.sha512(query.encode("utf-8")).hexdigest()
        logger.debug(f"Executing SPARQL query {query_first_line} ({query_hash}) against {endpoint_url}")
        start = datetime.now()
        sparql = SPARQLWrapper(endpoint_url.unicode_string(), agent=get_default_user_agent(), returnFormat=JSON)
        sparql.setQuery(query)
        sparql.setMethod(POST)
        resp = sparql.query().convert()
        lod_raw = resp.get("results", {}).get("bindings")
        logger.debug(
            f"Query ({query_hash}) execution finished! execution time : {(datetime.now() - start).total_seconds()}s, No. results: {len(lod_raw)}"  # noqa: E501
        )  # noqa: E501
        if logging.root.level <= logging.DEBUG:
            file_name = f"{start}_{query_hash}"
            cls.save_query(name=file_name, query=query)
            cls.save_results(name=file_name, lod=lod_raw)
        lod = []
        for d_raw in lod_raw:
            d = {key: record.get("value", None) for key, record in d_raw.items()}
            if d:
                lod.append(d)
        return lod

    @classmethod
    def check_availability_of_sparql_endpoint(cls, endpoint_url: HttpUrl):
        """
        Checks if the given endpoint is available.
        Uses the availability check proposed by Vandenbussche et al. see https://ceur-ws.org/Vol-1035/iswc2013_demo_21.pdf
        :param endpoint:
        :return: True if the endpoint is available. Otherwise, False
        """
        query = "ASK WHERE{ ?s ?p ?o . }"
        sparql = SPARQLWrapper(endpoint_url.unicode_string(), agent=get_default_user_agent(), returnFormat=JSON)
        sparql.setQuery(query)
        sparql.setMethod(POST)
        try:
            resp = sparql.query().convert()
            return resp.get("boolean", False)
        except Exception as e:
            logger.error(e)
        return False

    @classmethod
    def save_results(cls, name: str, lod: list[dict], path: Path | None = None) -> None:
        """
        store the results as json file
        :param lod:
        :param path:
        :return:
        """
        if path is None:
            path = cls._get_tmp_dir()
        path = path.joinpath(f"{name}.json")
        path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Saving results to {path}")
        with open(path, "w") as f:
            json.dump(lod, f, indent=2)

    @classmethod
    def save_query(cls, name: str, query: str, path: Path | None = None) -> None:
        """
        store the results as json file
        :param query:
        :param name:
        :param path:
        :return:
        """
        if path is None:
            path = cls._get_tmp_dir()
        path = path.joinpath(f"{name}.rq")
        path.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Saving query to {path}")
        with open(path, "w") as f:
            f.write(query)

    @classmethod
    def _get_tmp_dir(cls) -> Path:
        d = tempfile.gettempdir()
        path = Path(d).joinpath("WikibaseMigrator")
        path.mkdir(parents=True, exist_ok=True)
        return path

    @classmethod
    def add_namespace(cls, value: str, namespace: str) -> str:
        """
        Adds the namespace prefix to each value if needed
        :param value: value to which the prefix is added
        :param namespace: namespace to add
        :return: list of URLs
        """
        return value if value.startswith(namespace) else namespace + value

    @classmethod
    def remove_namespace(cls, value: str, namespace: str) -> str:
        """
        Removes the namespace prefix from the value if needed
        :param value:
        :param namespace:
        :return:
        """
        return value.removeprefix(namespace)

    @classmethod
    def get_sparql_uri(cls, value: str) -> str:
        return f"<{value}>"


class MediaWikiEndpoint:
    """
    Holds basic function to get data from a mediawiki api
    """

    @classmethod
    def get_supported_languages(cls, mediawiki_api_url: HttpUrl) -> dict[str, str]:
        """
        Get the languages supported by given MediaWiki
        :param mediawiki_api_url: api endpoint of the wiki
        :return: Mapping from language cot to the label
        """
        params = {
            "action": "query",
            "meta": "siteinfo",
            "siprop": "general|languages",
            "format": "json",
        }
        res = {}
        try:
            logger.debug(f"Querying supported languages from {mediawiki_api_url}")
            response = requests.get(mediawiki_api_url, params=params)
            data = response.json()
        except Exception as e:
            logger.error(e)
            data = {}
        for lang_record in data.get("query", {}).get("languages", []):
            code = lang_record.get("code")
            label = lang_record.get("*")
            if code:
                res[code] = label
        return res

    @classmethod
    def check_availability(cls, mediawiki_api_url: HttpUrl):
        """
        Check if the mediawiki api is available
        :param mediawiki_api_url:
        :return: True if the api is available, otherwise False
        """
        query = "action=query&titles=Main_Page&prop=revisions&rvprop=content&format=json"
        try:
            response = requests.get(f"{mediawiki_api_url.unicode_string()}?{query}")
            response.raise_for_status()
        except Exception:
            return False
        return True


class WikibaseEntityTypes(str, Enum):
    """
    wikibaseintegrator entity types
    """

    BASE_ENTITY = "base-entity"
    ITEM = "item"
    PROPERTY = "property"
    LEXEME = "lexeme"
    MEDIAINFO = "mediainfo"

    @classmethod
    def support_sitelinks(cls) -> list["WikibaseEntityTypes"]:
        """
        Returns list of Wikibase types which support sitelinks
        """
        return [WikibaseEntityTypes.ITEM]



"""A data model for quickstatements."""

import datetime
import logging
import webbrowser
from collections.abc import Iterable, Sequence
from enum import Enum
from typing import Annotated, Literal, get_args
from urllib.parse import quote

from pydantic import BaseModel, Field
from wikibaseintegrator.wbi_enums import WikibaseDatatype

logger = logging.getLogger(__name__)


def _safe_field(*, regex: str | None = None, **kwargs) -> Field:
    try:
        rv = Field(regex=regex, **kwargs)
    except TypeError:
        rv = Field(pattern=regex, **kwargs)
    return rv


class CalendarModels(str, Enum):
    """
    Wikidata Calendar Models
    """

    JULIAN = "http://www.wikidata.org/entity/Q1985786"
    GREGORIAN = "http://www.wikidata.org/entity/Q1985727"


class Qualifier(BaseModel):
    """
    A qualifier
    """

    type: str = "String"
    predicate: str = _safe_field(regex=r"^!?[PQS]\d+$")
    target: str

    def get_target(self) -> str:
        """Get the target wikidata identifier."""
        return self.target


class EntityQualifier(Qualifier):
    """A qualifier that points to Wikidata entity."""

    type: Literal["Entity"] = "Entity"
    target: str = _safe_field(regex=r"^[PQS]\d+$")


class DateQualifier(Qualifier):
    """A qualifier that points to a date string."""

    type: Literal["Date"] = "Date"

    @classmethod
    def point_in_time(
        cls,
        target: str | datetime.datetime | datetime.date,
        *,
        precision: int | None = None,
    ) -> "DateQualifier":
        """Get a qualifier for a point in time."""
        return cls(predicate="P585", target=prepare_date(target, precision=precision))

    @classmethod
    def start_time(
        cls,
        target: str | datetime.datetime | datetime.date,
        *,
        precision: int | None = None,
    ) -> "DateQualifier":
        """Get a qualifier for a start time."""
        return cls(predicate="P580", target=prepare_date(target, precision=precision))

    @classmethod
    def end_time(
        cls,
        target: str | datetime.datetime | datetime.date,
        *,
        precision: int | None = None,
    ) -> "DateQualifier":
        """Get a qualifier for an end time."""
        return cls(predicate="P582", target=prepare_date(target, precision=precision))

    @classmethod
    def retrieved(cls, namespace: Literal["P", "S"], precision: int | None = 11) -> "DateQualifier":
        """Get a qualifier for retrieving data now."""
        # FIXME this doesn't appear to work with higher granularity like 14
        now = datetime.datetime.now()
        return cls(predicate=f"{namespace}813", target=prepare_date(now, precision=precision))


def format_date(
    *,
    precision: int,
    year: int,
    month: int = 0,
    day: int = 0,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
) -> str:
    """Format the date in a way appropriate for quickstatements."""
    return f"+{year:04}-{month:02}-{day:02}T{hour:02}:{minute:02}:{second:02}Z/{precision}"


def prepare_date(
    target: str | datetime.datetime | datetime.date, *, precision: int | None = None, calendar: str | None = None
) -> str:
    """Prepare a date for quickstatements."""
    if isinstance(target, str):
        precision_def = f"/{precision}" if precision else ""
        if precision_def not in target:
            target += precision_def
        if calendar is not None and CalendarModels(calendar) is CalendarModels.JULIAN:
            target += "/J"
        return target
    if not isinstance(target, datetime.datetime | datetime.date):
        raise TypeError
    if precision is None:
        precision = 11
    if isinstance(target, datetime.date) and precision > 11:
        precision = 11
        logger.warning("Can not have higher precision on a datetime.date input than 11")
    # See section on precision in https://www.wikidata.org/wiki/Help:Dates#Precision
    if precision == 11:  # day precision
        return format_date(precision=precision, year=target.year, month=target.month, day=target.day)
    elif precision == 10:  # month precision
        return format_date(precision=precision, year=target.year, month=target.month)
    elif precision == 9:  # year precision
        return format_date(precision=precision, year=target.year)
    elif precision == 12:  # hour precision
        if not isinstance(target, datetime.datetime):
            raise RuntimeError
        return format_date(
            precision=precision,
            year=target.year,
            month=target.month,
            day=target.day,
            hour=target.hour,
        )
    elif precision == 13:  # minute precision
        if not isinstance(target, datetime.datetime):
            raise RuntimeError
        return format_date(
            precision=precision,
            year=target.year,
            month=target.month,
            day=target.day,
            hour=target.hour,
            minute=target.minute,
        )
    elif precision == 14:  # second precision
        if not isinstance(target, datetime.datetime):
            raise RuntimeError
        return format_date(
            precision=precision,
            year=target.year,
            month=target.month,
            day=target.day,
            hour=target.hour,
            minute=target.minute,
            second=target.second,
        )
    else:
        raise ValueError(f"Invalid precision: {precision}")
    # No precision case:
    # return f"+{target.isoformat()}Z"


class TextQualifier(Qualifier):
    """A qualifier that points to a string literal."""

    type: Literal["Text"] = "Text"

    def get_target(self) -> str:
        """Get the target text literal."""
        return f'"{self.target}"'


class TimeQualifier(DateQualifier):
    """A line whose target is a date/datetime."""

    precision: int | None = None
    calendar: str | None = None

    def get_target(self) -> str:
        """Get the date literal line."""
        return prepare_date(self.target, precision=self.precision, calendar=self.calendar)


class MonolingualTextQualifier(TextQualifier):
    """A qualifier whose target is a monolingual string."""

    language: str

    def get_target(self) -> str:
        """Get the target text literal."""
        return f'{self.language}:"{self.target}"'


class QuantityQualifier(Qualifier):
    type: Literal["Quantity"] = "Quantity"
    unit: str | None = _safe_field(regex=r"^[Q]\d+$")
    tolerance: str | None = None

    def get_target(self) -> str:
        """Get the target text literal."""
        res = f"{self.target}"
        if self.tolerance is not None:
            res += f"~{self.tolerance}"
        if self.unit is not None:
            unit = self.unit.replace("Q", "U")
            res += unit
        return res


class CreateLine(BaseModel):
    """A trivial model representing the CREATE line."""

    type: Literal["Create"] = "Create"

    def get_line(self, sep: str = "|") -> str:
        """Get the CREATE line."""
        return "CREATE"


class CreatePropertyLine(BaseModel):
    """A trivial model representing the CREATE line."""

    datatype: WikibaseDatatype
    type: Literal["Create"] = "Create"

    def get_line(self, sep: str = "|") -> str:
        """Get the CREATE line."""
        return f"CREATE_PROPERTY{sep}{self.datatype.value}"


class BaseLine(BaseModel):
    """A shared model for entity and text lines."""

    subject: str = _safe_field(regex=r"^(LAST)|(Q\d+)$")
    predicate: str = _safe_field(
        regex=r"^(P\d+)|([ADL][a-z]+)|(S\w+)$",
        description="""\
        The predicate can be one of two things:

        1. A Wikidata predicate, which starts with an upper case letter P, followed by a sequence of digits
        2. A combination of a single letter command code and an ISO639 language code.
           The single letter command codes can be:
           - ``L`` for label
           - ``A`` for alias (i.e., synonym)
           - ``D`` for description
           - ``S`` for sidelinks
           See Wikidata documentation at https://www.wikidata.org/w/index.php?title=Help:QuickStatements&\
section=6#Adding_labels,_aliases,_descriptions_and_sitelinks

        To do: add support for sitelinks.
        """,
    )
    qualifiers: list[Qualifier] = Field(default_factory=list)

    def get_target(self) -> str:
        """Get the target of the line."""
        return self.target

    def get_line(self, sep: str = "|") -> str:
        """Get the QuickStatement line as a string."""
        parts = [self.subject, self.predicate, self.get_target()]
        for qualifier in self.qualifiers:
            parts.append(qualifier.predicate)
            parts.append(qualifier.get_target())
        return sep.join(parts)


class EntityLine(BaseLine):
    """A line whose target is a string literal."""

    type: Literal["Entity"] = "Entity"
    target: str = _safe_field(regex=r"^[PQ]\d+$")


class TextLine(BaseLine):
    """A line whose target is a Wikidata entity."""

    type: Literal["Text"] = "Text"
    target: str

    def get_target(self) -> str:
        """Get the text literal line."""
        return f'"{self.target}"'


class MonolingualTextLine(TextLine):
    """A qualifier whose target is a monolingual string."""

    language: str

    def get_target(self) -> str:
        """Get the target text literal."""
        return f'{self.language}:"{self.target}"'


class QuantityLine(BaseLine):
    type: Literal["Quantity"] = "Quantity"
    target: str
    unit: str | None = _safe_field(regex=r"^[Q]\d+$")
    tolerance: str | None = None

    def get_target(self) -> str:
        """Get the target text literal."""
        res = f"{self.target}"
        if self.tolerance is not None:
            res += f"~{self.tolerance}"
        if self.unit is not None:
            unit = self.unit.replace("Q", "U")
            res += unit
        return res


class DateLine(BaseLine):
    """A line whose target is a date/datetime."""

    type: Literal["Date"] = "Date"
    target: datetime.datetime | datetime.date | str
    precision: int | None = None
    calendar: str | None = None

    def get_target(self) -> str:
        """Get the date literal line."""
        return prepare_date(self.target, precision=self.precision, calendar=self.calendar)


#: A union of the line types
Line = Annotated[CreateLine | EntityLine | TextLine | DateLine, Field(discriminator="type")]


def render_lines(lines: Iterable[Line], sep: str = "|", newline: str = "||") -> str:
    """Prepare QuickStatements line objects for sending to the API."""
    return newline.join(line.get_line(sep=sep) for line in lines)


def lines_to_url(lines: Iterable[Line]) -> str:
    """Prepare a URL for V1 of QuickStatements."""
    quoted_qs = quote(render_lines(lines), safe="")
    return f"https://quickstatements.toolforge.org/#/v1={quoted_qs}"


def lines_to_new_tab(lines: Iterable[Line]) -> bool:
    """Open a web browser on the host system.

    :param lines: QuickStatements lines
    :returns: If a web browser was successfully invoked (via :func:`webbrowser.open`)
    """
    lines = list(lines)
    if not lines:
        return False
    return webbrowser.open_new_tab(lines_to_url(lines))


def _unpack_annotated(t) -> Sequence[type]:
    return get_args(get_args(t)[0])


def write_json_schema():
    """Write a JSON schema."""
    import json

    import pydantic.schema

    schema = pydantic.schema.schema(
        [
            *_unpack_annotated(Qualifier),
            *_unpack_annotated(Line),
        ],
        title="QuickStatements",
        description="A data model representing lines in Quickstatements",
    )
    with open("schema.json", "w") as file:
        json.dump(schema, file, indent=2)


if __name__ == "__main__":
    write_json_schema()
